/*----------------------------------------------------------------------
  File    : bcload.c
  Contents: routines for loading Bayes classifiers and data
            (common to the X11 and the Windows program)
  Author  : Christian Borgelt
  History : 2000.02.18 file created from file xbcview.c
            2000.06.08 bug in function bc_load removed (ranges)
            2003.04.12 clustering result loading added
            2004.08.17 radial basis function networks added
            2007.01.10 usage of attribute maps changed
            2007.07.08 bug in loading RBF nets with numeric target fixed
            2013.08.28 learning vector quantization added
            2018.10.17 extended functions (attset, attmap) removed
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "bcload.h"

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
/* --- attribute information --- */
ATTSET  *attset  = NULL;        /* attribute set */
ATTMAP  *attmap  = NULL;        /* attribute map */
CCHAR   **names  = NULL;        /* attribute names */
RANGE   *ranges  = NULL;        /* ranges of attribute values */
ATTID   *map_fwd = NULL;        /* map indices to attset identifiers */
ATTID   *map_inv = NULL;        /* map attset identifiers to indices */
ATTID   attcnt   = 0;           /* number of (metric) attributes */
SELATT  selatt   =              /* attribute selection information */
  { 0, { 0.0, 0.0 }, 0, { 0.0, 0.0 } };

/* --- classifier and cluster information --- */
NBC     *nbc     = NULL;        /* naive Bayes classifier */
FBC     *fbc     = NULL;        /* full  Bayes classifier */
LVQNET  *lvq     = NULL;        /* learning vector quantization */
CLSET   *clset   = NULL;        /* cluster set (also as part of RBF) */
RBFNET  *rbf     = NULL;        /* radial basis function network */

/* --- display information --- */
BVNORM  *bvnorm  = NULL;        /* bivariate normal distributions */

/* --- data information --- */
TABLE   *table   = NULL;        /* data table */
DATAFMT datafmt  =              /* data format information */
  { 1, "\\n", " ,\\t", " \\t\\r", "?", "#" };

/* --- error message --- */
char    bc_msg[2*AS_MAXLEN+64]; /* buffer for an error message */

/* --- internal variables --- */
static TABREAD *tread = NULL;   /* table file reader */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

static void cleanup (int all)
{                               /* --- clean up after a load error */
  if (tread)   { trd_delete(tread, 1);  tread   = NULL; }
  if (table)   { tab_delete(table, 0);  table   = NULL; }
  if (!all) return;
  if (bvnorm)  { free(bvnorm);          bvnorm  = NULL; }
  if (rbf)     { rbf_delete(rbf, 1);    clset   = NULL; rbf = NULL; }
  if (fbc)     { fbc_delete(fbc, 0);    fbc     = NULL; }
  if (nbc)     { nbc_delete(nbc, 0);    nbc     = NULL; }
  if (lvq)     { lvq_delete(lvq);       lvq     = NULL; }
  if (clset)   { cls_delete(clset);     clset   = NULL; }
  if (attset)  { as_delete(attset);     attset  = NULL; }
  if (attmap)  { am_delete(attmap, 0);  attmap  = NULL; }
  if (ranges)  { free(ranges);          ranges  = NULL; }
  if (map_fwd) { free(map_fwd);         map_fwd = NULL; }
  if (names)   { free((void*)names);    names   = NULL; }
  attcnt = 0;                   /* free memory and clear variables */
}  /* cleanup() */

/*--------------------------------------------------------------------*/

static int error (int code, int all, const char *msg, SCANNER *scan)
{                               /* --- handle an error */
  if (scan) scn_delete(scan,1); /* delete the scanner */
  if (!msg) bc_msg[0] = 0;      /* note the error message */
  else strncpy(bc_msg, msg, sizeof(bc_msg));
  bc_msg[sizeof(bc_msg)-1] = 0; /* terminate the error message */
  cleanup(all);                 /* clear all/data variables */
  return code;                  /* return the error code */
}  /* error() */

/*--------------------------------------------------------------------*/

static ATTMAP* makemap (ATTSET *attset)
{                               /* --- create an attribute map */
  ATTID i;                      /* loop variable */
  ATT   *att;                   /* to traverse the attributes */

  assert(attset);               /* check the function argument */
  for (i = as_attcnt(attset); --i >= 0; ) {
    att = as_att(attset, i);    /* traverse the attributes */
    if (att_getdir(att) != DIR_IN) att_setmark(att, -1);
  }                             /* deselect non-input attributes */
  return am_create(attset, AM_MARKED, 1.0);
}  /* makemap() */              /* create an attribute map */

/*--------------------------------------------------------------------*/

void bc_clean (void)
{ cleanup(1); }                 /* --- clean up after an error */

/*--------------------------------------------------------------------*/

int bc_load (const char *fname)
{                               /* --- load classifier/clusters */
  ATTID   k, m, r;              /* loop variables and counters */
  VALID   c, clscnt;            /* number of classes/clusters */
  int     type;                 /* attribute type */
  SCANNER *scan;                /* scanner */
  ATT     *att;                 /* to traverse the attributes */
  MVNORM  *mvn;                 /* to traverse the normal distribs. */
  MATRIX  *mat;                 /* to traverse the matrices */
  DTINT   i;                    /* buffer for a min/max value */
  DTFLT   f;                    /* buffer for a min/max value */
  double  exp, dev, t;          /* expected value and std. deviation */

  /* --- parse domain descriptions --- */
  cleanup(1);                   /* clean up existing variables */
  attset = as_create("domains", att_delete);
  if (!attset) return error(E_NOMEM, 1, NULL, NULL);
  scan = scn_create();          /* create a scanner */
  if (!scan) return E_NOMEM;    /* for the classifier/cluster set */
  if (scn_open(scan, NULL, fname) != 0)
    return error(E_FOPEN, 1, fname, scan);
  if (as_parse(attset, scan, AT_ALL, 1) != 0)
    return error(E_PARSE, 1, fname, scan);

  /* --- load Bayes classifier/cluster set/RBF network --- */
  if (scn_token(scan) != T_ID){ /* check for a keyword */
    scn_error(scan, E_STREXP, "nbc");
    return error(E_PARSE, 1, fname, scan);
  }
  if (strcmp(scn_value(scan), "nbc") == 0) {
    nbc = nbc_parse(attset, scan); /* parse a naive Bayes classifier */
    if (!nbc || !scn_eof(scan, 1))
      return error(E_PARSE, 1, fname, scan);
    scn_delete(scan, 1);        /* delete the scanner */
    if (nbc_total(nbc) <= 0)    /* check for a positive case total */
      return error(E_CREATE, 1, NULL, NULL);
    clscnt = nbc_clscnt(nbc); } /* get the number of classes */
  else if (strcmp(scn_value(scan), "fbc") == 0) {
    fbc = fbc_parse(attset, scan); /* parse a full Bayes classifier */
    if (!fbc || !scn_eof(scan, 1))
      return error(E_PARSE, 1, fname, scan);
    scn_delete(scan, 1);        /* delete the scanner */
    if (fbc_total(fbc) <= 0)    /* check for a positive case total */
      return error(E_CREATE, 1, NULL, NULL);
    clscnt = fbc_clscnt(fbc); } /* get the number of classes */
  else if ((strcmp(scn_value(scan), "lvq")    == 0)
  ||       (strcmp(scn_value(scan), "lvqnet") == 0)) {
    attmap = makemap(attset);   /* create an attribute map */
    if (!attmap) return error(E_NOMEM, 1, NULL, scan);
    lvq = lvq_parse(scan, am_inpcnt(attmap), 1);
    if (!lvq || !scn_eof(scan, 1)) /* parse an LVQ network */
      return error(E_PARSE, 1, fname, scan);
    scn_delete(scan, 1);        /* delete the scanner */
    lvq_unscale(lvq);           /* remove all attribute scaling */
    clscnt = lvq_cnt(lvq); }    /* get number of reference vectors */
  else if ((strcmp(scn_value(scan), "clset")    == 0)
  ||       (strcmp(scn_value(scan), "clusters") == 0)) {
    attmap = makemap(attset);   /* create an attribute map */
    if (!attmap) return error(E_NOMEM, 1, NULL, scan);
    clset = cls_parse(scan, am_inpcnt(attmap), 1);
    if (!clset || !scn_eof(scan, 1)) /* parse a cluster set */
      return error(E_PARSE, 1, fname, scan);
    scn_delete(scan, 1);        /* delete the scanner */
    clscnt = cls_clscnt(clset); /* get the number of clusters */
    cls_unscale(clset);         /* remove all attribute scaling */
    cls_setup(clset); }         /* and set up the cluster set */
  else if ((strcmp(scn_value(scan), "rbf")    == 0)
  ||       (strcmp(scn_value(scan), "rbfnet") == 0)) {
    attmap = makemap(attset);   /* create an attribute map */
    if (!attmap) return error(E_NOMEM, 1, NULL, scan);
    rbf = rbf_parse(scan, NULL, 0, 1);
    if (!rbf || !scn_eof(scan, 1)) /* parse an RBF network */
      return error(E_PARSE, 1, fname, scan);
    scn_delete(scan, 1);        /* delete the scanner */
    clset  = rbf_clset(rbf);    /* get the underlying cluster set */
    clscnt = cls_clscnt(clset); /* get the number of clusters */
    cls_unscale(clset);         /* remove all attribute scaling */
    cls_setup(clset);           /* and set up the cluster set */
    m = cls_inpcnt(clset); }    /* get the number of attributes */
  else {                        /* if keyword is not known */
    scn_error(scan, E_STREXP, "nbc");
    return error(E_PARSE, 1, fname, scan);
  }                             /* abort with failure */

  /* --- create the attribute vectors --- */
  m = as_attcnt(attset);        /* get the number of attributes */
  for (k = attcnt = 0; k < m; k++) {
    type = att_type(as_att(attset, k));
    if ((type == AT_INT) || (type == AT_FLT)) attcnt++;
  }                             /* count the numeric attributes */
  if (attcnt < 1)               /* and check for at least one */
    return error(E_ATTCNT, 1, fname, NULL);
  bvnorm  = (BVNORM*)malloc(   (size_t)clscnt *sizeof(BVNORM));
  names   = (CCHAR**)malloc(   (size_t)attcnt *sizeof(CCHAR*));
  ranges  = (RANGE*) malloc(   (size_t)attcnt *sizeof(RANGE));
  map_fwd = (ATTID*) malloc(2 *(size_t)attcnt *sizeof(ATTID));
  if (!bvnorm || !names || !ranges || !map_fwd)
    return error(E_NOMEM, 1, fname, NULL);
  map_inv = map_fwd +attcnt;    /* create the attribute arrays */

  /* --- build classifier and attribute information --- */
  selatt.h_att = selatt.v_att = -1; /* init. attribute indices */
  for (k = attcnt = 0; k < m; k++) {
    att  = as_att(attset, k);   /* traverse the attributes */
    type = att_type(att);       /* and get the attribute type */
    if ((type != AT_INT) && (type != AT_FLT))
      continue;                 /* skip non-numeric attributes */
    if      (selatt.h_att < 0)  /* note the first  numeric attribute */
      selatt.h_att = attcnt;    /* for the horizontal direction */
    else if (selatt.v_att < 0)  /* note the second numeric attribute */
      selatt.v_att = attcnt;    /* for the vertical direction */
    names  [attcnt] = att_name(att);
    map_fwd[attcnt] = k;        /* note the attribute name */
    map_inv[k]      = attcnt;   /* and the attribute identifier */
    if (type == AT_INT) {       /* if attribute is integer valued */
      i = att_valmin(att)->i;   /* get and set minimum value */
      ranges[attcnt].min = (i <= DTINT_MIN) ? -INFINITY : (double)i;
      i = att_valmax(att)->i;   /* get and set maximum value */
      ranges[attcnt].max = (i >= DTINT_MAX) ?  INFINITY : (double)i; }
    else {                      /* if attribute is real valued */
      f = att_valmin(att)->f;   /* get and set minimum value */
      ranges[attcnt].min = (f <= DTFLT_MIN) ? -INFINITY : (double)f;
      f = att_valmax(att)->f;   /* get and set maximum value */
      ranges[attcnt].max = (f >= DTFLT_MAX) ?  INFINITY : (double)f;
    }                           /* (set initial range of values ) */
    for (c = clscnt; --c >= 0;){/* traverse the normal distributions */
      if      (nbc) {           /* if naive Bayes classifier */
        exp =      nbc_exp(nbc, c, k);
        dev = sqrt(nbc_var(nbc, c, k)); }
      else if (fbc) {           /* if full Bayes classifier */
        mvn = fbc_mvnorm(fbc, c);
        exp =      mvn_exp(mvn, k);
        dev = sqrt(mvn_var(mvn, k)); }
      else if (lvq) {           /* if learning vector quantization */
        r   = am_off(attmap, k);
        exp = lvq_center(lvq, c)[r];
        dev = sqrt(lvq_var(lvq, c)); }
      else {                    /* if cluster set */
        r   = am_off(attmap, k);
        exp = cls_center(clset, c)[r];
        mat = cls_covmat(clset, c);
        dev = sqrt((mat) ? mat_get(mat, r, r) : cls_isovar(clset, c));
      }                         /* get exp. value and std. deviation */
      t = exp -3 *dev;          /* update minimum value of range */
      if (t < ranges[attcnt].min) ranges[attcnt].min = t;
      t = exp +3 *dev;          /* update maximum value of range */
      if (t > ranges[attcnt].max) ranges[attcnt].max = t;
    }
    if (ranges[attcnt].min >= ranges[attcnt].max) {
      ranges[attcnt].min = 0; ranges[attcnt].max = 1; }
    attcnt++;                   /* check and adapt final range */
  }                             /* and go to the next attribute */
  if (selatt.v_att < 0) selatt.v_att = selatt.h_att;
  selatt.h_rng = ranges[selatt.h_att];
  selatt.v_rng = ranges[selatt.v_att];
  return 0;                     /* get ranges of selected attributes */
}  /* bc_load() */              /* and return 'ok' */

/*--------------------------------------------------------------------*/

int bc_data (const char *fname)
{                               /* --- load data table */
  int   r;                      /* result of a function call */
  int   m;                      /* table file read mode */
  ATTID k;                      /* loop variable for attributes */
  ATT   *att;                   /* class attribute */

  if (!attset) return E_DATA;   /* check for an attribute set */
  if (table) {                  /* delete an existing table */
    tab_delete(table, 0); table = NULL; }
  for (k = as_attcnt(attset); --k >= 0; )
    att_setmark(as_att(attset, k), 1);
  if (nbc || fbc) {             /* mark all attributes and if class. */
    att = as_att(attset, (nbc) ? nbc_trgid(nbc) : fbc_trgid(fbc));
    att_setmark(att, 0);        /* get the class attribute, */
    att_inst(att)->n = NV_NOM;  /* mark it as optional, and */
  }                             /* set its value to unknown */
  /* With these preparations and the flag AS_MARKED in the call to  */
  /* as_read() the class is read only if it is present in the data. */
  /* Otherwise the class is set to unknown in all tuples.           */

  table = tab_create("table", attset, tpl_delete);
  if (!table)  return error(E_NOMEM, 0, NULL, NULL);
  tread = trd_create();         /* create a table file reader */
  if (!tread)  return error(E_NOMEM, 0, NULL, NULL);
  trd_allchs(tread, datafmt.recseps, datafmt.fldseps,
                    datafmt.blanks,  datafmt.nullchs, datafmt.comment);
  if (trd_open(tread, NULL, fname) != 0) /* open the input file */
    return error(E_FOPEN, 0, trd_name(tread), NULL);
  m = ((datafmt.names) ? AS_ATT : AS_DFLT) | AS_MARKED;
  r = tab_read(table, tread, m);/* read the table file */
  if (r < 0) return error(-r, 0, tab_errmsg(table, NULL, 0), NULL);
  trd_delete(tread, 1);         /* delete the table file reader */
  tread = NULL;                 /* and clear the corresp. variable */
  return 0;                     /* return 'ok' */
}  /* bc_data() */

/*--------------------------------------------------------------------*/

VALID bc_bvnorm (void)
{                               /* --- init. normal distributions */
  VALID  i, k;                  /* loop variables */
  VALID  clscnt;                /* number of classes */
  ATTID  x, y;                  /* attribute identifiers */
  MVNORM *mvn;                  /* to traverse the normal distribs. */
  double *ctr;                  /* to traverse the cluster centers */
  MATRIX *cov;                  /* to traverse the covariance mats. */
  double t0, t1, t2;            /* temporary buffers */

  x = map_fwd[selatt.h_att];    /* get the attribute identifiers */
  y = map_fwd[selatt.v_att];    /* of the selected attributes */
  if (nbc) {                    /* if naive Bayes classifier, */
    clscnt = nbc_clscnt(nbc);   /* get the number of classes */
    for (i = k = 0; i < clscnt; i++) {
      k |= bvn_init(bvnorm +i, nbc_prior(nbc, i),
                    nbc_exp(nbc, i, x), nbc_exp(nbc, i, y),
                    nbc_var(nbc, i, x), nbc_var(nbc, i, y), 0);
    } }                         /* create bivariate normal dists. */
  else if (fbc) {               /* if full Bayes classifier, */
    clscnt = fbc_clscnt(fbc);   /* get the number of classes */
    for (i = k = 0; i < clscnt; i++) {
      mvn = fbc_mvnorm(fbc, i); /* traverse the normal distributions */
      k |= bvn_init(bvnorm +i, fbc_prior(fbc, i),
                    mvn_exp(mvn, x), mvn_exp(mvn, y),
                    mvn_var(mvn, x), mvn_var(mvn, y),
                    (x != y) ? mvn_cov(mvn, x, y) : 0);
    } }                         /* create bivariate normal dists. */
  else if (lvq) {               /* if learning vector quantization */
    clscnt = lvq_cnt(lvq);      /* get the number of ref. vectors */
    for (i = k = 0; i < clscnt; i++) {
      ctr = lvq_center(lvq, i); /* traverse the reference vectors */
      t0  = lvq_var   (lvq, i); /* and get center and variance */
      k  |= bvn_init(bvnorm +i, 1.0, ctr[x], ctr[y], t0, t0, 0);
    } }                         /* create bivariate normal dists. */
  else {                        /* if cluster set */
    x = am_off(attmap, x);      /* get the rows/columns */
    y = am_off(attmap, y);      /* associated with the attributes */
    clscnt = cls_clscnt(clset); /* get the number of classes */
    for (i = k = 0; i < clscnt; i++) {
      ctr = cls_center(clset, i);  /* traverse the clusters */
      cov = cls_covmat(clset, i);  /* and get their parameters */
      t0  = mat_get(cov, x, x);    /* get the variances */
      t1  = mat_get(cov, y, y);    /* and the covariance */
      t2  = (x != y) ? mat_get(cov, x, y) : 0;
      k  |= bvn_init(bvnorm +i, cls_weight(clset, i),
                     ctr[x], ctr[y], t0, t1, t2);
    }                           /* create a bivariate normal dist. */
  }                             /* for each of the clusters */
  return (k != 0) ? -1 : clscnt;/* return the initialization status */
}  /* bc_bvnorm() */
