/*----------------------------------------------------------------------
  File    : bcload.h
  Contents: routines for loading Bayes classifiers and data
            (common to the X11 and the Windows program)
  Author  : Christian Borgelt
  History : 2000.02.18 file created from file xbcview.c
            2000.11.30 cluster sets replaced by a full Bayes classifier
            2001.05.13 definition of E_ATTYPE removed
            2003.04.12 clustering result loading added
            2004.05.15 function bc_bvnorm added
            2004.08.17 radial basis function networks added
            2013.08.28 learning vector quantization added
----------------------------------------------------------------------*/
#ifndef __BCLOAD__
#define __BCLOAD__
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_WRITE
#define AS_WRITE
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#include "table.h"
#ifndef NBC_PARSE
#define NBC_PARSE
#endif
#include "nbayes.h"
#ifndef FBC_PARSE
#define FBC_PARSE
#endif
#include "fbayes.h"
#ifndef LVQ_PARSE
#define LVQ_PARSE
#endif
#ifndef LVQ_EXTFN
#define LVQ_EXTFN
#endif
#include "lvq.h"
#ifndef CLS_PARSE
#define CLS_PARSE
#endif
#ifndef CLS_EXTFN
#define CLS_EXTFN
#endif
#include "cluster.h"
#ifndef RBF_PARSE
#define RBF_PARSE
#endif
#ifndef RBF_EXTFN
#define RBF_EXTFN
#endif
#include "rbf.h"
#include "bvnorm.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- data formats --- */
#define FR_ATTS        0        /* attribute names in first record */
#define FR_DATA        1        /* data tuple in first record */
#define FR_COMMENT     2        /* comment in first record */

#define FMTCHRLEN     80        /* length of format character strings */

/* --- error codes --- */
#define E_PARSE      (-5)       /* initialization failed */
#define E_CREATE     (-6)       /* cannot create classifier */
#define E_ATTCNT     (-7)       /* too few numeric attributes */
#define E_DATA       (-8)       /* classifier must be loaded first */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* -- data format description -- */
  int    names;                 /* column names in first record */
  char   recseps[FMTCHRLEN+1];  /* record separators */
  char   fldseps[FMTCHRLEN+1];  /* field separators */
  char   blanks [FMTCHRLEN+1];  /* blank characters */
  char   nullchs[FMTCHRLEN+1];  /* null value characters */
  char   comment[FMTCHRLEN+1];  /* comment characters */
} DATAFMT;                      /* (data format description) */

typedef struct {                /* --- range of values --- */
  double min, max;              /* minimum and maximum value */
} RANGE;                        /* (range of values) */

typedef struct {                /* -- selected attributes -- */
  ATTID  h_att;                 /* attribute for horizontal direction */
  RANGE  h_rng;                 /* horizontal range of values */
  ATTID  v_att;                 /* attribute for vertical   direction */
  RANGE  v_rng;                 /* vertical range of values */
} SELATT;                       /* (selected attributes) */

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
/* --- attribute information --- */
extern ATTSET  *attset;         /* attribute set */
extern ATTMAP  *attmap;         /* attribute map */
extern CCHAR   **names;         /* attribute names */
extern RANGE   *ranges;         /* ranges of attribute values */
extern ATTID   *map_fwd;        /* map indices to attset identifiers */
extern ATTID   *map_inv;        /* map attset identifiers to indices */
extern ATTID   attcnt;          /* number of (metric) attributes */
extern SELATT  selatt;          /* attribute selection information */

/* --- classifier and cluster information --- */
extern NBC     *nbc;            /* naive Bayes classifier */
extern FBC     *fbc;            /* full  Bayes classifier */
extern LVQNET  *lvq;            /* learning vector quantization */
extern CLSET   *clset;          /* cluster set */
extern RBFNET  *rbf;            /* radial basis function network */

/* --- display information --- */
extern BVNORM  *bvnorm;         /* bivariate normal distributions */

/* --- data information --- */
extern TABLE   *table;          /* data table */
extern DATAFMT datafmt;         /* data format description */

/* --- error message --- */
extern char    bc_msg[];        /* buffer for an (error) message */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern void  bc_clean  (void);
extern int   bc_load   (const char *fname);
extern int   bc_data   (const char *fname);
extern VALID bc_bvnorm (void);

#endif
